{ ####################################################################################### }
{ ##                                                                                   ## }
{ ## Main_Unit                                                                         ## }
{ ##                                                                                   ## }
{ ## Main form for db_transmitter                                                      ## }
{ ##                                                                                   ## }
{ ## Copyright (C) 2018-2019  : Dr. Jürgen Abel                                        ## }
{ ## Email                    : juergen@mve.info                                       ## }
{ ## Internet                 : www.seismometer.info                                   ## }
{ ##                                                                                   ## }
{ ## This program is free software: you can redistribute it and/or modify              ## }
{ ## it under the terms of the GNU Lesser General Public License as published by       ## }
{ ## the Free Software Foundation, either version 3 of the License, or                 ## }
{ ## (at your option) any later version with the following modification:               ## }
{ ##                                                                                   ## }
{ ## As a special exception, the copyright holders of this library give you            ## }
{ ## permission to link this library with independent modules to produce an            ## }
{ ## executable, regardless of the license terms of these independent modules, and     ## }
{ ## to copy and distribute the resulting executable under terms of your choice,       ## }
{ ## provided that you also meet, for each linked independent module, the terms        ## }
{ ## and conditions of the license of that module. An independent module is a          ## }
{ ## module which is not derived from or based on this library. If you modify          ## }
{ ## this library, you may extend this exception to your version of the library,       ## }
{ ## but you are not obligated to do so. If you do not wish to do so, delete this      ## }
{ ## exception statement from your version.                                            ## }
{ ##                                                                                   ## }
{ ## This program is distributed in the hope that it will be useful,                   ## }
{ ## but WITHOUT ANY WARRANTY; without even the implied warranty of                    ## }
{ ## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                     ## }
{ ## GNU General Public License for more details.                                      ## }
{ ##                                                                                   ## }
{ ## You should have received a copy of the GNU Lesser General Public License          ## }
{ ## COPYING.LGPL.txt along with this program.                                         ## }
{ ## If not, see <https://www.gnu.org/licenses/>.                                      ## }
{ ##                                                                                   ## }
{ ####################################################################################### }


Unit Main_Unit;

{$mode objfpc}{$H+}

{$INCLUDE project_defines.inc}

Interface

Uses
  CThreads,
  Classes,
  SysUtils,
  Forms,
  Controls,
  Graphics,
  Dialogs,
  StdCtrls,
  ExtCtrls,
  ComCtrls,
  TAGraph,
  TASeries,
  ADS1262_Unit,
  TACustomSeries,
  TATools,
  PQConnection,
  SQLDB;

Const
  {$IFDEF USE_INTERRUPT}
  {$ELSE}
  DATA_INTERVAL_MS                  = 2;
  {$ENDIF}
  LP_SPAN_INIT                      = 20;
  DATA_ARRY_SIZE                    = 1024 * 1024;
  CHART_EXTEND_Y                    = 0.0005;
  CHART_EXTEND_Y_V                  = 5;
  VALUES_PER_HOUR                   = 60 * 60;
  RMS_COLOR                         = $2222AA;
  RMS_BAR_WIDTH_PERCENT             = 100;
  DATA_COLOR                        = $FF4444;
  DATA_PIN_WIDTH                    = 3;
  POSITION_COLOR                    = $6666EE;
  POSITION_WIDTH                    = 11;
  IMAGE_FILE_NAME                   = 'images/daily_image.jpg';
  HOUR_CHART_N                      = 24;
  HOUR_CHART_FONT_SIZE              = 6;
  STATION_NAME                      = 'NORF_00';
  CURRENT_TIME_FORMAT_00_S          = 'Live Data  -  Station Name: ' + STATION_NAME + '  -  Date: ';
  CURRENT_TIME_FORMAT_01_S          = '  -  Time: ';
  SQR_2                             = sqrt (2);
  BIT_N_RANGE                       = 100;
  PROGRAM_START_CLEAR_DB_S          = 'Clear db';
  PROGRAM_START_CREATE_DB_S         = 'Create db';
  PROGRAM_START_S                   = 'Start program';
  {$IFDEF USE_DB_LOCALHOST}
  SQL_HOSTNAME                      = '127.0.0.1';
  {$ELSE}
  SQL_HOSTNAME                      = '192.168.0.99';
  {$ENDIF}
  SQL_DATABASENAME                  = 'seismometer_db';
  SQL_USERNAME                      = 'user';
  SQL_PASSWORD                      = '123456';
  SQL_POSTGRES_USERNAME             = 'postgres';
  SQL_POSTGRES_PASSWORD             = '123456';
  SQL_RESTART_SEQUENCE_TXT          = 'ALTER SEQUENCE adc_values_id_seq RESTART WITH 0;';
  SQL_END_TRANSATION_TXT            = 'END TRANSACTION;';
  SQL_BEGIN_TRANSATION_TXT          = 'BEGIN TRANSACTION;';
  SQL_VACUUM_FULL_TXT               = 'VACUUM FULL;';
  SQL_VACUUM_ANALYZE_TXT            = 'VACUUM ANALYZE;';
  SQL_ADC_VALUES_RECORD_N           = 400 * 60 * 60 * 24 * 31;
  SQL_ADC_VALUES_RECORD_AT_ONCE_N   = 1000;
  SQL_ADC_VALUES_TABLE_NAME_TXT     = 'adc_values';
  SQL_ADC_VALUES_ID_TXT             = 'id';
  SQL_ADC_VALUES_TIME_STAMP_TXT     = 'time_stamp';
  SQL_ADC_VALUES_ADC_VALUE_TXT      = 'adc_value';
  SQL_ADC_VALUES_TRUNCATE_TXT       = 'TRUNCATE ' + SQL_ADC_VALUES_TABLE_NAME_TXT + ';';
  SQL_ADC_VALUES_INSERT_DUMMIES_TXT = 'INSERT INTO ' + SQL_ADC_VALUES_TABLE_NAME_TXT + ' (' + SQL_ADC_VALUES_TIME_STAMP_TXT + ', ' + SQL_ADC_VALUES_ADC_VALUE_TXT + ') VALUES (CURRENT_TIMESTAMP , 0);';
  SQL_ADC_VALUES_UPDATE_TXT         = 'UPDATE ' + SQL_ADC_VALUES_TABLE_NAME_TXT + ' SET ' + SQL_ADC_VALUES_TIME_STAMP_TXT + ' = :' + SQL_ADC_VALUES_TIME_STAMP_TXT + ', ' + SQL_ADC_VALUES_ADC_VALUE_TXT + ' = :' + SQL_ADC_VALUES_ADC_VALUE_TXT + ' WHERE ' + SQL_ADC_VALUES_ID_TXT + ' = :' + SQL_ADC_VALUES_ID_TXT + ';';
  SQL_ID_VALUES_TABLE_NAME_TXT      = 'id_values';
  SQL_ID_VALUES_ID_TXT              = 'id';
  SQL_ID_VALUES_ID_VALUE_TXT        = 0;
  SQL_ID_VALUES_ADC_VALUES_ID_TXT   = 'adc_values_id';
  SQL_ID_VALUES_UPDATE_TXT          = 'UPDATE ' + SQL_ID_VALUES_TABLE_NAME_TXT + ' SET ' + SQL_ID_VALUES_ADC_VALUES_ID_TXT + ' = :' + SQL_ID_VALUES_ADC_VALUES_ID_TXT + ' WHERE ' + SQL_ADC_VALUES_ID_TXT + ' = :' + SQL_ADC_VALUES_ID_TXT + ';';

Type
  T_Data_X_A = Array [0 .. DATA_ARRY_SIZE - 1] Of TDateTime;
  T_Data_Y_A = Array [0 .. DATA_ARRY_SIZE - 1] Of Int_32;

  { ####################################################################################### }
  { ## T_Timer_Thread                                                                    ## }
  { ####################################################################################### }
  T_Timer_Thread = Class(TThread)
  Private
      M_ADS1262 :      T_ADS1262;
      M_Data_X_A :     T_Data_X_A;
      M_Data_Y_A :     T_Data_Y_A;
      M_Data_A_Index : Integer;
      M_Count :        Int_64;
  Public
      Property Data_X_A : T_Data_X_A Read M_Data_X_A;
      Property Data_Y_A : T_Data_Y_A Read M_Data_Y_A;
      Property Data_A_Index : Integer Read M_Data_A_Index;
      Property Count : Int_64 Read M_Count;
      Property ADS1262 : T_ADS1262 Read M_ADS1262;
      Constructor Create (F_Suspended : Boolean);
      Destructor Destroy; Override;
      Procedure Execute; Override;
  End; { T_Timer_Thread }

  { ####################################################################################### }
  { ## TMain_F                                                                           ## }
  { ####################################################################################### }

  { TMain_F }

  TMain_F = Class(TForm)
      ADC_Values_Q :           TSQLQuery;
      ID_Values_Q :            TSQLQuery;
      Close_B :                TButton;
      Input_Timer_Sec_E :      TEdit;
      Input_Timer_Sec_L :      TLabel;
      Postgres_C :             TPQConnection;
      Postgres_T :             TSQLTransaction;
      Input_OFCAL_Value_E :    TEdit;
      Input_OFCAL_Value_L :    TLabel;
      Output_1_P :             TPanel;
      Output_2_P :             TPanel;
      Output_Bits_E :          TEdit;
      Output_Bits_L :          TLabel;
      Output_Counter_E :       TEdit;
      Output_Counter_L :       TLabel;
      Output_Freq_E :          TEdit;
      Output_Freq_L :          TLabel;
      Output_NPS_E :           TEdit;
      Input_Sensivity_Bits_E : TEdit;
      Output_NPS_L :           TLabel;
      Input_Sensivity_Bits_L : TLabel;
      Start_Stop_B :           TButton;
      Output_0_P :             TPanel;
      Chart_T :                TTimer;
      Clear_DB_T :             TTimer;
      Status_P :               TPanel;
      Status_L :               TLabel;
      Procedure Chart_TTimer (Sender : TObject);
      Procedure Clear_DB_TTimer (Sender : TObject);
      Procedure Close_BClick (Sender : TObject);
      Procedure FormCreate (Sender : TObject);
      Procedure FormDestroy (Sender : TObject);
      Procedure FormShow (Sender : TObject);
      Procedure Input_Timer_Sec_EEditingDone (Sender : TObject);
      Procedure Input_OFCAL_Value_EEditingDone (Sender : TObject);
      Procedure Input_Sensivity_Bits_EEditingDone (Sender : TObject);
      Procedure Start_Stop_BClick (Sender : TObject);
  Protected
      M_Timer_Thread :      T_Timer_Thread;
      M_Data_A_Index_0 :    Integer;
      M_Data_A_Index_1 :    Integer;
      M_id :                Int_64;
      M_Start_Time :        TDateTime;
      M_Old_Data_Hour :     Integer;
      M_Old_Data_Minute :   Integer;
      M_Old_Data_Second :   Integer;
      M_Last_Saved_Minute : Integer;
      Procedure Show_Status (F_Status : String);
      Function Validate_Array_Index (F_Index : Integer; F_Array_Size : Integer) : Integer;
      Procedure Output_Current_Data (F_Measurement_N : Integer; F_Zero_Crossings_N : Integer; F_End_Time : TDateTime);
  Public
  End; { TMain_F }

Var
  Main_F : TMain_F;

Implementation

{$R *.frm}

Uses
  Math,
  DateUtils,
  TAChartUtils;

{ ####################################################################################### }
{ ## T_Timer_Thread                                                                    ## }
{ ####################################################################################### }

{ --------------------------------------------------------------------------------------- }
Constructor T_Timer_Thread.Create (F_Suspended : Boolean);
  { Create timer thread                                                                     }
  { --------------------------------------------------------------------------------------- }
{$IFDEF DEBUG_ERROR}
Var
  Register_Block : T_Register_Block;
{$ENDIF}
Begin { T_Timer_Thread.Create }
  Inherited Create (F_Suspended);

  FreeOnTerminate := FALSE;

  M_ADS1262 := T_ADS1262.Create ();
  M_ADS1262.Set_Continuous_Conversion ();
  M_ADS1262.Start_Conversion ();
  Sleep (ADS1262_WAIT_TIME_MS_COMMAND);

  {$IFDEF DEBUG_ERROR}
  M_ADS1262.Read_All_Registers (Register_Block);
  {$ENDIF}

  FillByte (M_Data_X_A [0], DATA_ARRY_SIZE * SizeOf (TDateTime), $00);
  FillByte (M_Data_Y_A [0], DATA_ARRY_SIZE * SizeOf (Int_32), $00);

  M_Data_A_Index := 0;
  M_Count        := 0;
End; { T_Timer_Thread.Create }


{ --------------------------------------------------------------------------------------- }
Destructor T_Timer_Thread.Destroy;
  { Free data                                                                               }
  { --------------------------------------------------------------------------------------- }
Begin { T_Timer_Thread.Destroy }
  M_ADS1262.Free;

  Inherited Destroy;
End; { T_Timer_Thread.Destroy }


{ --------------------------------------------------------------------------------------- }
Procedure T_Timer_Thread.Execute;
{ Execute thread                                                                          }
{ --------------------------------------------------------------------------------------- }
Begin { T_Timer_Thread.Execute }
  While (Terminated = FALSE) Do
    Begin { While }
      {$IFDEF USE_INTERRUPT}
      M_ADS1262.DRDY_Pin.Wait_For_Interrupt (GPIO_TIMEOUT_INTERRUPT_INFINITE);
      {$ELSE}
      Sleep (DATA_INTERVAL_MS);
      {$ENDIF}

      M_Data_X_A[M_Data_A_Index] := Now;
      M_Data_Y_A[M_Data_A_Index] := M_ADS1262.ADC1_Data;

      Inc (M_Count);
      Inc (M_Data_A_Index);
      If M_Data_A_Index >= DATA_ARRY_SIZE Then
        Begin { then }
          M_Data_A_Index := 0;
        End; { then }
    End;     { While }
End; { T_Timer_Thread.Execute }


{ ####################################################################################### }
{ ## TMain_F                                                                           ## }
{ ####################################################################################### }

{#####################################################################################}
{ Build form                                                                          }
{#####################################################################################}

{ --------------------------------------------------------------------------------------- }
Procedure TMain_F.FormCreate (Sender : TObject);
{ Create main from                                                                        }
{ --------------------------------------------------------------------------------------- }
Begin { TMain_F.FormCreate }
  M_Old_Data_Hour     := 0;
  M_Old_Data_Minute   := 0;
  M_Old_Data_Second   := 0;
  M_Last_Saved_Minute := 0;
  M_id                := 0;

  M_Timer_Thread := T_Timer_Thread.Create (TRUE);
  M_Timer_Thread.ADS1262.Set_OFCAL (StrToInt (Input_OFCAL_Value_E.Text));
End; { TMain_F.FormCreate }


{ --------------------------------------------------------------------------------------- }
Procedure TMain_F.FormDestroy (Sender : TObject);
{ Free data                                                                               }
{ --------------------------------------------------------------------------------------- }
Begin { TMain_F.FormDestroy }
  Chart_T.Enabled := FALSE;

  M_Timer_Thread.Terminate;
  Sleep (100);
  M_Timer_Thread.Free;

  Postgres_C.Close;
End; { TMain_F.FormDestroy }


{ --------------------------------------------------------------------------------------- }
Procedure TMain_F.FormShow (Sender : TObject);
{ Show main form                                                                          }
{ --------------------------------------------------------------------------------------- }
Begin { TMain_F.FormShow }
  {$IFDEF INIT_DB_AT_START}
  Clear_DB_T.Enabled   := TRUE;
  {$ELSE}
  ADC_Values_Q.Active  := FALSE;
  ID_Values_Q.Active   := FALSE;
  Postgres_T.Active    := FALSE;
  Postgres_C.Connected := FALSE;

  Postgres_C.Transaction  := Postgres_T;
  Postgres_C.HostName     := SQL_HOSTNAME;
  Postgres_C.DatabaseName := SQL_DATABASENAME;

  Postgres_C.UserName  := SQL_USERNAME;
  Postgres_C.Password  := SQL_PASSWORD;
  Postgres_C.Connected := TRUE;

  ADC_Values_Q.Close;
  ADC_Values_Q.SQL.Clear;
  ADC_Values_Q.SQL.Add (SQL_ADC_VALUES_UPDATE_TXT);

  ID_Values_Q.Close;
  ID_Values_Q.SQL.Clear;
  ID_Values_Q.SQL.Add (SQL_ID_VALUES_UPDATE_TXT);

  M_Data_A_Index_0 := 0;
  M_Data_A_Index_1 := 0;
  M_Start_Time     := Now;

  M_Timer_Thread.Start;

  Chart_T.Interval := StrToInt (Input_Timer_Sec_E.Text);
  Chart_T.Enabled  := TRUE;
  {$ENDIF}
End; { TMain_F.FormShow }


{#####################################################################################}
{ Buttons and edit components                                                         }
{#####################################################################################}

{ --------------------------------------------------------------------------------------- }
Procedure TMain_F.Show_Status (F_Status : String);
{ Show status and update form                                                             }
{ --------------------------------------------------------------------------------------- }
Begin { TMain_F.Show_Status }
  Status_L.Caption := F_Status;

  Update;
  Application.ProcessMessages;
End; { TMain_F.Show_Status }


{ --------------------------------------------------------------------------------------- }
Procedure TMain_F.Output_Current_Data (F_Measurement_N : Integer; F_Zero_Crossings_N : Integer; F_End_Time : TDateTime);
{ --------------------------------------------------------------------------------------- }
Var
  I :         Integer;
  Bit_N_Max : Double;

Begin { TMain_F.Output_Current_Data }
  Bit_N_Max := 0;
  For I := 0 To BIT_N_RANGE - 1 Do
    Begin { For }
      Bit_N_Max := Max (Bit_N_Max, Log2 (Max (Abs (M_Timer_Thread.Data_Y_A [(M_Data_A_Index_1 - 1 - I) mod DATA_ARRY_SIZE]), 1)));
    End; { For }

  Output_NPS_E.Text     := FormatFloat ('#,##0.0', Double (F_Measurement_N) / (Double (MilliSecondSpan (M_Start_Time, F_End_Time)) / 1000));
  Output_Counter_E.Text := FormatFloat ('#,##0', M_Timer_Thread.Data_A_Index);
  Output_Bits_E.Text    := FormatFloat ('#,##0.0', Bit_N_Max);
  Output_Freq_E.Text    := FormatFloat ('#,##0.0', F_Zero_Crossings_N / 2 / Double (MilliSecondSpan (M_Start_Time, F_End_Time)) * 1000);
  M_Start_Time          := F_End_Time;

  Update;
  Application.ProcessMessages;
End; { TMain_F.Output_Current_Data }


{ --------------------------------------------------------------------------------------- }
Procedure TMain_F.Close_BClick (Sender : TObject);
{ Close button pressed                                                                    }
{ --------------------------------------------------------------------------------------- }
Begin { TMain_F.Close_BClick }
  Close;
End; { TMain_F.Close_BClick }


{ --------------------------------------------------------------------------------------- }
Procedure TMain_F.Start_Stop_BClick (Sender : TObject);
{ Start/Stop button pressed                                                               }
{ --------------------------------------------------------------------------------------- }
Begin { TMain_F.Start_Stop_BClick }
  Chart_T.Enabled := not Chart_T.Enabled;
End; { TMain_F.Start_Stop_BClick }


{ --------------------------------------------------------------------------------------- }
Procedure TMain_F.Input_Sensivity_Bits_EEditingDone (Sender : TObject);
{ --------------------------------------------------------------------------------------- }
Begin { TMain_F.Input_Sensivity_Bits_EEditingDone }
    Try
      M_Timer_Thread.ADS1262.Sensivity := StrToInt (Input_Sensivity_Bits_E.Text);
    Except
      M_Timer_Thread.ADS1262.Sensivity := ADS1262_INITIAL_SENSIVITY;
    End; { Try }
End; { TMain_F.Input_Sensivity_Bits_EEditingDone }


{ --------------------------------------------------------------------------------------- }
Procedure TMain_F.Input_OFCAL_Value_EEditingDone (Sender : TObject);
{ --------------------------------------------------------------------------------------- }
Begin { TMain_F.Input_OFCAL_Value_EEditingDone }
    Try
      M_Timer_Thread.ADS1262.Set_OFCAL (StrToInt (Input_OFCAL_Value_E.Text));
    Except
      M_Timer_Thread.ADS1262.Set_OFCAL (ADS1262_INITIAL_OFCAL);
    End; { Try }
End; { TMain_F.Input_OFCAL_Value_EEditingDone }


{ --------------------------------------------------------------------------------------- }
Procedure TMain_F.Input_Timer_Sec_EEditingDone (Sender : TObject);
{ --------------------------------------------------------------------------------------- }
Begin { TMain_F.Input_Chart_Timer_Sec_EEditingDone }
  Chart_T.Interval := StrToInt (Input_Timer_Sec_E.Text);
End; { TMain_F.Input_Chart_Timer_Sec_EEditingDone }


{#####################################################################################}
{ Utilities                                                                           }
{#####################################################################################}

{ --------------------------------------------------------------------------------------- }
Function TMain_F.Validate_Array_Index (F_Index : Integer; F_Array_Size : Integer) : Integer;
  { Validate index into array                                                               }
  { --------------------------------------------------------------------------------------- }
Begin { TMain_F.Validate_Array_Index }
  If F_Index >= 0 Then
    Begin { then }
      Result := F_Index mod F_Array_Size;
    End { then }
  Else
    Begin { else  }
      Result := ((F_Index mod F_Array_Size) + F_Array_Size) mod F_Array_Size;
    End; { else  }
End; { TMain_F.Validate_Array_Index }


{#####################################################################################}
{ DB data                                                                             }
{#####################################################################################}

{ --------------------------------------------------------------------------------------- }
Procedure TMain_F.Clear_DB_TTimer (Sender : TObject);
{ Clear and initialize db                                                                 }
{ --------------------------------------------------------------------------------------- }
Var
  I :        Integer;
  {$IFDEF INSERT_DUMMIES_AT_START}
  End_Time : TDateTime;
  {$ENDIF}
  S :        String;

Begin { TMain_F.Clear_DB_TTimer }
  Clear_DB_T.Enabled := FALSE;

  ADC_Values_Q.Active     := FALSE;
  ID_Values_Q.Active      := FALSE;
  Postgres_T.Active       := FALSE;
  Postgres_C.Connected    := FALSE;
  Postgres_C.Transaction  := Postgres_T;
  Postgres_C.HostName     := SQL_HOSTNAME;
  Postgres_C.DatabaseName := SQL_DATABASENAME;

  {$IFDEF CLEAR_DB_AT_START}
  Show_Status (PROGRAM_START_CLEAR_DB_S);

  Postgres_C.UserName  := SQL_POSTGRES_USERNAME;
  Postgres_C.Password  := SQL_POSTGRES_PASSWORD;
  Postgres_C.Connected := TRUE;
  Postgres_C.ExecuteDirect (SQL_END_TRANSATION_TXT);
  Postgres_C.ExecuteDirect (SQL_ADC_VALUES_TRUNCATE_TXT);
  Postgres_C.ExecuteDirect (SQL_RESTART_SEQUENCE_TXT);
  Postgres_C.ExecuteDirect (SQL_VACUUM_FULL_TXT);
  Postgres_C.ExecuteDirect (SQL_VACUUM_ANALYZE_TXT);
  Postgres_T.Commit;
  Postgres_T.EndTransaction;
  Postgres_C.Connected := FALSE;
  {$ENDIF}

  Postgres_C.UserName  := SQL_USERNAME;
  Postgres_C.Password  := SQL_PASSWORD;
  Postgres_C.Connected := TRUE;

  ADC_Values_Q.Close;
  S := '';
  For I := 0 To SQL_ADC_VALUES_RECORD_AT_ONCE_N - 1 Do
    Begin { For }
      S := S + SQL_ADC_VALUES_INSERT_DUMMIES_TXT + Chr ($0D);
    End; { For }

  ID_Values_Q.Close;
  ID_Values_Q.SQL.Clear;
  ID_Values_Q.SQL.Add (SQL_ID_VALUES_UPDATE_TXT);

  {$IFDEF INSERT_DUMMIES_AT_START}
  Show_Status (PROGRAM_START_CREATE_DB_S);

  M_Start_Time := Now;
  Postgres_T.StartTransaction;
  For I := 0 To (SQL_ADC_VALUES_RECORD_N div SQL_ADC_VALUES_RECORD_AT_ONCE_N) - 1 Do
    Begin { For }
      { Add several rows at once }
      ADC_Values_Q.SQL.Clear;
      ADC_Values_Q.SQL.Add (S);
      ADC_Values_Q.Prepare;
      ADC_Values_Q.ExecSQL;

      End_Time := Now;
      Show_Status ('Record counter: ' + FormatFloat ('#,##0', I * SQL_ADC_VALUES_RECORD_AT_ONCE_N) + ' --- Record rate : ' + FormatFloat ('#,##0.0', Double (SQL_ADC_VALUES_RECORD_AT_ONCE_N) / (Double (MilliSecondSpan (M_Start_Time, End_Time)) / 1000)));
      M_Start_Time := End_Time;
    End; { For }

  { Save last id }
  ID_Values_Q.Prepare;
  ID_Values_Q.Params.ParamByName (SQL_ID_VALUES_ID_TXT).AsLargeInt            := SQL_ID_VALUES_ID_VALUE_TXT;
  ID_Values_Q.Params.ParamByName (SQL_ID_VALUES_ADC_VALUES_ID_TXT).AsLargeInt := 0;
  ID_Values_Q.ExecSQL;

  Postgres_T.Commit;
  {$ENDIF}

  ADC_Values_Q.Close;
  ADC_Values_Q.SQL.Clear;
  ADC_Values_Q.SQL.Add (SQL_ADC_VALUES_UPDATE_TXT);

  Show_Status ('');

  M_Data_A_Index_0 := 0;
  M_Data_A_Index_1 := 0;
  M_Start_Time     := Now;

  {$IFDEF UPDATE_VALUES_AFTER_START}
  M_Timer_Thread.Start;

  Chart_T.Interval := StrToInt (Input_Timer_Sec_E.Text);
  Chart_T.Enabled  := TRUE;
  {$ELSE}
  Close;
  {$ENDIF}
End; { TMain_F.Clear_DB_TTimer }


{ --------------------------------------------------------------------------------------- }
Procedure TMain_F.Chart_TTimer (Sender : TObject);
{ Repaint chart                                                                           }
{ --------------------------------------------------------------------------------------- }
Var
  I :                Integer;
  Start_Index :      Integer;
  End_Index :        Integer;
  Voltage :          Double;
  Old_Voltage :      Double;
  Zero_Crossings_C : Integer;
  Additional_Index : Integer;

Begin { TMain_F.Chart_TTimer }
  Postgres_T.StartTransaction;

  M_Data_A_Index_1 := M_Timer_Thread.Data_A_Index;
  Zero_Crossings_C := 0;
  Additional_Index := 0;

  If M_Data_A_Index_1 < M_Data_A_Index_0 Then
    Begin { then }
      Start_Index := M_Data_A_Index_0;
      End_Index   := DATA_ARRY_SIZE - 1;
      Voltage     := 0;
      Old_Voltage := Voltage;
      I           := Start_Index;
      While I <= End_Index Do
        Begin { While }
          Voltage := M_Timer_Thread.ADS1262.Calculate_Voltage (M_Timer_Thread.Data_Y_A [I]);

          ADC_Values_Q.Prepare;
          ADC_Values_Q.Params.ParamByName (SQL_ADC_VALUES_ID_TXT).AsLargeInt         := M_id;
          ADC_Values_Q.Params.ParamByName (SQL_ADC_VALUES_TIME_STAMP_TXT).AsDateTime := M_Timer_Thread.Data_X_A [I];
          ADC_Values_Q.Params.ParamByName (SQL_ADC_VALUES_ADC_VALUE_TXT).AsInteger   := M_Timer_Thread.Data_Y_A [I];
          ADC_Values_Q.ExecSQL;

          If (I > Start_Index) and ((Old_Voltage < 0) and (Voltage > 0) or (Old_Voltage > 0) and (Voltage < 0)) Then
            Begin { then }
              Inc (Zero_Crossings_C);
            End; { then }

          Old_Voltage := Voltage;
          M_id        := Validate_Array_Index (M_id + 1, SQL_ADC_VALUES_RECORD_N);
          Inc (I);
        End; { While }

      Additional_Index := End_Index - Start_Index;
      M_Data_A_Index_0 := 0;
    End; { then }

  Start_Index := M_Data_A_Index_0;
  End_Index   := M_Data_A_Index_1 - 1;
  Voltage     := 0;
  Old_Voltage := Voltage;
  I           := Start_Index;
  While I <= End_Index Do
    Begin { While }
      Voltage := M_Timer_Thread.ADS1262.Calculate_Voltage (M_Timer_Thread.Data_Y_A [I]);

      ADC_Values_Q.Prepare;
      ADC_Values_Q.Params.ParamByName (SQL_ADC_VALUES_ID_TXT).AsLargeInt         := M_id;
      ADC_Values_Q.Params.ParamByName (SQL_ADC_VALUES_TIME_STAMP_TXT).AsDateTime := M_Timer_Thread.Data_X_A [I];
      ADC_Values_Q.Params.ParamByName (SQL_ADC_VALUES_ADC_VALUE_TXT).AsInteger   := M_Timer_Thread.Data_Y_A [I];
      ADC_Values_Q.ExecSQL;

      If (I > Start_Index) and ((Old_Voltage < 0) and (Voltage > 0) or (Old_Voltage > 0) and (Voltage < 0)) Then
        Begin { then }
          Inc (Zero_Crossings_C);
        End; { then }

      Old_Voltage := Voltage;
      M_id        := Validate_Array_Index (M_id + 1, SQL_ADC_VALUES_RECORD_N);
      Inc (I);
    End; { While }

  ID_Values_Q.Prepare;
  ID_Values_Q.Params.ParamByName (SQL_ID_VALUES_ID_TXT).AsLargeInt            := SQL_ID_VALUES_ID_VALUE_TXT;
  ID_Values_Q.Params.ParamByName (SQL_ID_VALUES_ADC_VALUES_ID_TXT).AsLargeInt := M_id;
  ID_Values_Q.ExecSQL;

  Postgres_T.Commit;

  Output_Current_Data (End_Index - Start_Index + Additional_Index, Zero_Crossings_C, M_Timer_Thread.Data_X_A [I - 1]);

  M_Data_A_Index_0 := M_Data_A_Index_1;
End; { TMain_F.Chart_TTimer }


End.
